@@ -1,13 +1,9 @@
-//! # Usage
-//!
-//! ```shell
-//! apk-downloader -h
-//! ```
-//!
 //! # List Sources
 //!
 //! A few distinct lists of APKs are used.  AndroidRank compiles the most popular apps available on
-//! the Google Play Store.
+//! the Google Play Store.  You can also specify a CSV file which lists the apps to download.  If
+//! you have a simple file with one app ID per line, you can just treat it as a CSV with a single
+//! field.
 //!
 //! # Download Sources
 //!
@@ -21,7 +17,6 @@
 #[macro_use]
 extern crate clap;
 
-use clap::{App, Arg};
 use futures_util::StreamExt;
 use gpapi::error::{Error as GpapiError, ErrorKind};
 use gpapi::Gpapi;
@@ -33,53 +28,44 @@ use std::path::Path;
 use std::rc::Rc;
 use std::time::Duration;
 use thirtyfour::prelude::*;
+use tokio::time::{sleep, Duration as TokioDuration};
 
-arg_enum! {
-    #[derive(Debug)]
-    pub enum ListSource {
-        AndroidRank,
-        ISDi,
-    }
-}
-arg_enum! {
-    pub enum DownloadSource {
-        APKPure,
-        GooglePlay,
-    }
+mod cli;
+use cli::{ListSource, DownloadSource};
+
+async fn fetch_android_rank_list() -> Result<Vec<String>, Box<dyn Error>> {
+    let resp = reqwest::get("https://www.androidrank.org/applist.csv")
+        .await?
+        .error_for_status()?
+        .text()
+        .await?;
+
+    Ok(parse_csv_text(resp, 1))
 }
 
-async fn fetch_list(source: &ListSource) -> Result<Vec<String>, Box<dyn Error>> {
-    let resp = match source {
-        &ListSource::AndroidRank => reqwest::get("https://www.androidrank.org/applist.csv"),
-        &ListSource::ISDi => reqwest::get("https://raw.githubusercontent.com/stopipv/isdi/master/static_data/app-flags.csv"),
-    }.await?
-     .error_for_status()?
-     .text()
-     .await?;
+fn fetch_csv_list(csv: &str, field: usize) -> Result<Vec<String>, Box<dyn Error>> {
+    Ok(parse_csv_text(fs::read_to_string(csv)?, field))
+}
 
-    Ok(resp.split("\n").filter_map(|l| {
+fn parse_csv_text(text: String, field: usize) -> Vec<String> {
+    let field = field - 1;
+    text.split("\n").filter_map(|l| {
         let entry = l.trim();
         let mut entry_vec = entry.split(",").collect::<Vec<&str>>();
-        if entry_vec.len() > 2 {
-            match source {
-                &ListSource::AndroidRank => Some(String::from(entry_vec.remove(0))),
-                &ListSource::ISDi => {
-                    let app_id = entry_vec.remove(0);
-                    match entry_vec.remove(0) {
-                        "playstore" => Some(String::from(app_id)),
-                        _ => None,
-                    }
-                },
-            }
+        if entry_vec.len() > field && !(entry_vec.len() == 1 && entry_vec[0].len() == 0) {
+            Some(String::from(entry_vec.remove(field)))
         } else {
             None
         }
-    }).collect())
+    }).collect()
 }
 
-async fn download_apps_from_google_play(app_ids: Vec<String>, parallel: usize, username: &str, password: &str, outpath: &str) {
+async fn download_apps_from_google_play(app_ids: Vec<String>, parallel: usize, sleep_duration: u64, username: &str, password: &str, outpath: &str) {
     let mut gpa = Gpapi::new("en_US", "UTC", "hero2lte");
-    gpa.login(username, password).await.expect("Could not log in to google play");
+    if let Err(_) = gpa.login(username, password).await {
+        println!("Could not log in to Google Play.  Please check your credentials and try again later.");
+        std::process::exit(1);
+    }
     let gpa = Rc::new(gpa);
 
     futures_util::stream::iter(
@@ -87,6 +73,9 @@ async fn download_apps_from_google_play(app_ids: Vec<String>, parallel: usize, u
             let gpa = Rc::clone(&gpa);
             async move {
                 println!("Downloading {}...", app_id);
+                if sleep_duration > 0 {
+                    sleep(TokioDuration::from_millis(sleep_duration)).await;
+                }
                 match gpa.download(&app_id, None, &Path::new(outpath)).await {
                     Ok(_) => Ok(()),
                     Err(err) if matches!(err.kind(), ErrorKind::FileExists) => {
@@ -119,19 +108,19 @@ async fn download_apps_from_google_play(app_ids: Vec<String>, parallel: usize, u
     ).buffer_unordered(parallel).collect::<Vec<Result<(), GpapiError>>>().await;
 }
 
-async fn download_apps_from_apkpure(app_ids: Vec<String>, parallel: usize, outpath: &str) -> WebDriverResult<()> {
+async fn download_apps_from_apkpure(app_ids: Vec<String>, parallel: usize, sleep_duration: u64, outpath: &str) -> WebDriverResult<()> {
     let fetches = futures_util::stream::iter(
         app_ids.into_iter().map(|app_id| {
             async move {
-                match download_single_app(&app_id, outpath).await {
+                match download_single_app(&app_id, sleep_duration, outpath).await {
                     Ok(res_tuple) => futures_util::future::ready(Some(res_tuple)),
                     Err(_) => {
                         println!("An error has occurred attempting to download {}.  Retry #1...", app_id);
-                        match download_single_app(&app_id, outpath).await {
+                        match download_single_app(&app_id, sleep_duration, outpath).await {
                             Ok(res_tuple) => futures_util::future::ready(Some(res_tuple)),
                             Err(_) => {
                                 println!("An error has occurred attempting to download {}.  Retry #2...", app_id);
-                                match download_single_app(&app_id, outpath).await {
+                                match download_single_app(&app_id, sleep_duration, outpath).await {
                                     Ok(res_tuple) => futures_util::future::ready(Some(res_tuple)),
                                     Err(_) => {
                                         println!("An error has occurred attempting to download {}.  Aborting.", app_id);
@@ -164,8 +153,11 @@ async fn download_apps_from_apkpure(app_ids: Vec<String>, parallel: usize, outpa
     Ok(())
 }
 
-async fn download_single_app(app_id: &str, outpath: &str) -> WebDriverResult<(String, String, String)> {
+async fn download_single_app(app_id: &str, sleep_duration: u64, outpath: &str) -> WebDriverResult<(String, String, String)> {
     println!("Downloading {}...", app_id);
+    if sleep_duration > 0 {
+        sleep(TokioDuration::from_millis(sleep_duration)).await;
+    }
     let app_url = format!("https://apkpure.com/a/{}/download?from=details", app_id);
     let mut caps = DesiredCapabilities::chrome();
     let filepath = format!("{}", Path::new(outpath).join(app_id.clone()).to_str().unwrap());
@@ -178,8 +170,8 @@ async fn download_single_app(app_id: &str, outpath: &str) -> WebDriverResult<(St
         Ok(driver) => driver,
         Err(_) => panic!("chromedriver must be running on port 4444")
     };
-    let delay = Duration::new(10, 0);
-    driver.set_implicit_wait_timeout(delay).await?;
+    let sleep_duration = Duration::new(10, 0);
+    driver.set_implicit_wait_timeout(sleep_duration).await?;
     driver.get(app_url).await?;
     let elem_result = driver.find_element(By::Css("span.file")).await?;
     let re = Regex::new(r" \([0-9.]+ MB\)$").unwrap();
@@ -191,85 +183,49 @@ async fn download_single_app(app_id: &str, outpath: &str) -> WebDriverResult<(St
 
 #[tokio::main]
 async fn main() -> WebDriverResult<()> {
-    let matches = App::new("APK Downloader")
-        .author("William Budington <bill@eff.org>")
-        .about("Downloads APKs from various sources")
-        .usage("apk-downloader <-a app_name | -l list_source> [--download-source download_source] [--parallel parallel] OUTPUT ")
-        .arg(
-            Arg::with_name("list_source")
-                .help("Source of the apps list")
-                .short("l")
-                .long("list-source")
-                .takes_value(true)
-                .possible_values(&ListSource::variants()))
-        .arg(
-            Arg::with_name("app_name")
-                .help("Provide the name of an app directly")
-                .short("a")
-                .long("app-name")
-                .takes_value(true)
-                .conflicts_with("list_source")
-                .required_unless("list_source"))
-        .arg(
-            Arg::with_name("download_source")
-                .help("Where to download the APKs from")
-                .short("d")
-                .long("download-source")
-                .default_value("APKPure")
-                .takes_value(true)
-                .possible_values(&DownloadSource::variants())
-                .required(false))
-        .arg(
-            Arg::with_name("google_username")
-                .help("Google Username (required if download source is Google Play)")
-                .short("u")
-                .long("username")
-                .takes_value(true)
-                .required_if("download_source", "GooglePlay"))
-        .arg(
-            Arg::with_name("google_password")
-                .help("Google App Password (required if download source is Google Play)")
-                .short("p")
-                .long("password")
-                .takes_value(true)
-                .required_if("download_source", "GooglePlay"))
-        .arg(
-            Arg::with_name("parallel")
-                .help("The number of parallel APK fetches to run at a time")
-                .short("r")
-                .long("parallel")
-                .takes_value(true)
-                .default_value("4")
-                .required(false))
-        .arg(Arg::with_name("OUTPUT")
-            .help("An absolute path to store output files")
-            .required(true)
-            .index(1))
-        .get_matches();
+    let matches = cli::app().get_matches();
 
     let download_source = value_t!(matches.value_of("download_source"), DownloadSource).unwrap();
     let parallel = value_t!(matches, "parallel", usize).unwrap();
-    let outpath = matches.value_of("OUTPUT").unwrap();
+    let sleep_duration = value_t!(matches, "sleep_duration", u64).unwrap();
+    let outpath = matches.value_of("OUTPATH").unwrap();
     if !Path::new(&outpath).is_dir() {
-        println!("{}\n\nOUTPUT is not a valid directory", matches.usage());
+        println!("{}\n\nOUTPATH is not a valid directory", matches.usage());
         std::process::exit(1);
     };
     let list = match matches.value_of("app_name") {
         Some(app_name) => vec![app_name.to_string()],
         None => {
             let list_source = value_t!(matches.value_of("list_source"), ListSource).unwrap();
-            fetch_list(&list_source).await.unwrap()
+            match list_source {
+                ListSource::AndroidRank => fetch_android_rank_list().await.unwrap(),
+                ListSource::CSV => {
+                    let csv = matches.value_of("csv").unwrap();
+                    let field = value_t!(matches, "field", usize).unwrap();
+                    if field < 1 {
+                        println!("{}\n\nField must be 1 or greater", matches.usage());
+                        std::process::exit(1);
+                    }
+                    match fetch_csv_list(csv, field) {
+                        Ok(csv_list) => csv_list,
+                        Err(err) => {
+                            println!("{}\n\n{:?}", matches.usage(), err);
+                            std::process::exit(1);
+                        }
+                    }
+                }
+            }
         }
     };
 
     match download_source {
         DownloadSource::APKPure => {
-            download_apps_from_apkpure(list, parallel, outpath).await.unwrap();
+            download_apps_from_apkpure(list, parallel, sleep_duration, outpath).await.unwrap();
         },
         DownloadSource::GooglePlay => {
             let username = matches.value_of("google_username").unwrap();
             let password = matches.value_of("google_password").unwrap();
-            download_apps_from_google_play(list, parallel, username, password, outpath).await;
+            download_apps_from_google_play(list, parallel, sleep_duration, username, password, outpath).await;
         },
     }
     Ok(())
